/*
 * Copyright (C) 1997 WIDE Project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by WIDE Project and
 *	its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/*
 * $Id: draw.c,v 1.33.2.59 1997/12/26 05:52:35 itojun Exp $
 */

#include "mgp.h"

/* image/ *.[ch] */
extern void bestVisual __P((Display *, int, Image *, Visual **, unsigned int *));

/* state associated with the window - how should we treat this? */
static struct ctrl *bg_image = NULL;
static struct ctrl *bg_image_prev = NULL;


static short ctable[128] = {
	0x2121, 0x2121, 0x2121, 0x2121, 0x2121, 0x2121, 0x2121, 0x2121,
	0x2121, 0x2121, 0x2121, 0x2121, 0x2121, 0x2121, 0x2121, 0x2121,
	0x2121, 0x2121, 0x2121, 0x2121, 0x2121, 0x2121, 0x2121, 0x2121,
	0x2121, 0x2121, 0x2121, 0x2121, 0x2121, 0x2121, 0x2121, 0x2121,
	0x2121, 0x212a, 0x216d, 0x2174, 0x2170, 0x2173, 0x2175, 0x216c,
	0x214a, 0x214b, 0x2176, 0x215c, 0x2124, 0x215d, 0x2125, 0x213f,
	0x2330, 0x2331, 0x2332, 0x2333, 0x2334, 0x2335, 0x2336, 0x2337,
	0x2338, 0x2339, 0x2127, 0x2128, 0x2152, 0x2161, 0x2153, 0x2129,
	0x2177, 0x2341, 0x2342, 0x2343, 0x2344, 0x2345, 0x2346, 0x2347,
	0x2348, 0x2349, 0x234a, 0x234b, 0x234c, 0x234d, 0x234e, 0x234f,
	0x2350, 0x2351, 0x2352, 0x2353, 0x2354, 0x2355, 0x2356, 0x2357,
	0x2358, 0x2359, 0x235a, 0x214e, 0x2140, 0x214f, 0x2130, 0x2132,
	0x212e, 0x2361, 0x2362, 0x2363, 0x2364, 0x2365, 0x2366, 0x2367,
	0x2368, 0x2369, 0x236a, 0x236b, 0x236c, 0x236d, 0x236e, 0x236f,
	0x2370, 0x2371, 0x2372, 0x2373, 0x2374, 0x2375, 0x2376, 0x2377,
	0x2378, 0x2379, 0x237a, 0x2150, 0x2143, 0x2151, 0x2141, 0x2121,
};

#define	COMPLEX_BGIMAGE \
    (bg_image_prev				\
     && ((bg_image_prev->ct_op == CTL_BIMAGE)	\
       || bg_image_prev->ct_op == CTL_BGRAD))

#define	PM_POSY(height)	((window_height - (height)) / 2)

static void process_direc __P((struct render_state *, int *));

static int set_position __P((struct render_state *));
static void draw_line_start __P((struct render_state *));
static void draw_line_itemsize __P((struct render_state *, u_int));
static void draw_line_output __P((struct render_state *, u_char *));
static void draw_line_end __P((struct render_state *));
static void cutin __P((struct render_state *, Drawable, Drawable, int, int, int));
#if 0
static void shrink __P((u_char *, u_int));
#endif

static void draw_string __P((struct render_state *, u_char *));
#ifdef VFLIB
static u_int draw_onechar_vf __P((struct render_state *, u_int, u_int, u_int,
	u_int, u_int));
#endif
static u_int draw_onechar_x __P((struct render_state *, u_int, u_int,
	u_int));

static void back_gradation __P((struct render_state *, struct ctrl_grad *));
static void image_load __P((struct render_state *, char *, int, int, int, int, int, int));
static void process_icon __P((struct render_state *, struct ctrl *));
static void draw_bar __P((struct render_state *, struct ctrl *));
static void process_system __P((struct render_state *, struct ctrl *));

#ifndef HAVE_USLEEP
#if TIME_WITH_SYS_TIME
# include <sys/time.h>
#  include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif
void
usleep(usec)
	u_int usec;
{
#ifdef HAVE_POLL
	(void)poll(NULL, 0, usec/1000);
#else /* HAVE_POLL */
	struct timeval tv;
	tv.tv_sec  = usec / 1000000;
	tv.tv_usec = usec % 1000000;
	(void)select(0, NULL, NULL, NULL, &tv);
#endif /* HAVE_POLL */
}
#endif /* HAVE_USLEEP */

/*
 * state management.
 */
void
state_goto(state, page)
	struct render_state *state;
	u_int page;
{
	purgechild(state->page);
	state->page = page;
	state->line = 0;
	state->cp = NULL;
	state->phase = P_NONE;
}

void
state_next(state)
	struct render_state *state;
{
	switch (state->phase) {
	case P_NONE:
		fprintf(stderr, "internal error\n");
		break;
	case P_DEFAULT:
		if (state->cp)
			state->cp = state->cp->ct_next;
		if (!state->cp) {
			state->cp = page_control[state->page][state->line];
			state->phase = P_PAGE;
		}
		break;
	case P_PAGE:
		if (state->cp)
			state->cp = state->cp->ct_next;
		if (!state->cp) {
			state->line++;
			state->cp = NULL;
			state->phase = P_NONE;
			state_init(state);
		}
		break;

	case P_END:
		/*nothing*/
		break;
	}

	/* next page */
	if (page_attribute[state->page].pg_linenum < state->line) {
		purgechild(state->page);
		if (state->page < maxpage) {
			state->phase = P_NONE;
			state->page++;
			state->line = 0;
			state_newpage(state);
			state_init(state);
		} else
			state->phase = P_END;
	}
}

void
state_init(state)
	struct render_state *state;
{
	if (!state->cp || state->phase == P_NONE) {
#if 0
		if (!(page_attribute[state->page].pg_flag & PGFLAG_NODEF)) {
			state->cp = default_control[state->line];
			state->phase = P_DEFAULT;
		} else
#endif
		{
			state->cp = page_control[state->page][state->line];
			state->phase = P_PAGE;
		}
	}
}

void
state_newpage(state)
	struct render_state *state;
{
	state->ypos = 0;
	state->have_mark = 0;
}

/*
 * page management.
 */
void
draw_page(state)
	struct render_state *state;
{
	u_int end_line;
	u_int save_pl_mode;
	int pause;

	assert(state);

again:
	/* initialize the state, if required. */
	if (!state->cp || state->phase == P_NONE) {
		state_newpage(state);
		state_init(state);
	}

	end_line = page_attribute[state->page].pg_linenum;
	save_pl_mode = pl_mode;
	pl_off();

	while (1) {
		switch (state->phase) {
		case P_NONE:
			fprintf(stderr, "internal error\n");
			exit(-1);
		case P_DEFAULT:
		case P_PAGE:
			pause = 0;
			if (state->cp)
				process_direc(state, &pause);
			if (pause)
				goto done;
			break;
		case P_END:
			goto done;
		}
		XFlush(display);
		state_next(state);
	}

done:
	if (save_pl_mode)
		pl_on();
	if (pg_mode)
		pg_on(state->page, (state->line > end_line));
}

static void
process_direc(state, seenpause)
	struct render_state *state;
	int *seenpause;
{
	struct ctrl *cp;

	if (seenpause)
		*seenpause = 0;
	cp = state->cp;

	switch(cp->ct_op) {
	case CTL_SIZE:
		char_size = window_height * cp->cti_value / 100;
		break;

	case CTL_VGAP:
		vert_gap = cp->cti_value;
		break;

	case CTL_HGAP:
		horiz_gap = cp->cti_value;
		break;

	case CTL_QUALITY:
		if (!quality_flag)
			b_quality = cp->cti_value;
		break;

	case CTL_PAUSE:
		if (seenpause)
			*seenpause = 1;
		break;

	case CTL_AGAIN:
		if (state->have_mark)
			state->ypos = state->mark_ypos;
		state->have_mark = 0;
		break;

	case CTL_FORE:
		fore_color = cp->ctl_value;
		XSetForeground(display, gcfore, fore_color);
		XSetForeground(display, gcfb, fore_color);
		break;

	case CTL_BACK:
		bg_image = cp;	/*update later*/
		break;

	case CTL_CCOLOR:
		ctrl_color = cp->ctl_value;
		break;

	case CTL_CENTER:
		state->align = AL_CENTER;
		break;

	case CTL_LEFT:
		state->align = AL_LEFT;
		break;

	case CTL_LEFTFILL:
		state->align = AL_LEFTFILL0;
		break;

	case CTL_RIGHT:
		state->align = AL_RIGHT;
		break;

	case CTL_CONT:
		break;

#ifdef VFLIB
	case CTL_VFONT:
		vfc_setfont(cp->ctc_value);
		break;
#endif /*VFLIB*/

	case CTL_XFONT:
		strcpy(state->xfont, cp->ctc_value);
		break;

	case CTL_BAR:
		draw_bar(state, cp);
		break;

	case CTL_IMAGE:
		image_load(state, cp->ctm_fname, cp->ctm_numcolor,
			cp->ctm_ximagesize, cp->ctm_yimagesize, 0,
			cp->ctm_zoomflag, 0);
		break;

	case CTL_BIMAGE:
		if (mgp_flag & FL_BIMAGE)
			break;
		bg_image = cp;	/*update later*/
		break;

	case CTL_BGRAD:
		if (mgp_flag & FL_BIMAGE)
			break;
		bg_image = cp;	/*update later*/
		break;

	case CTL_LCUTIN:
		state->special = SP_LCUTIN;
		break;

	case CTL_RCUTIN:
		state->special = SP_RCUTIN;
		break;

	case CTL_SHRINK:
		state->special = SP_SHRINK;
		break;

	case CTL_PREFIX:
		state->curprefix = cp->ctc_value;
		break;

	case CTL_TABPREFIX:
		state->tabprefix = cp->ctc_value;
		break;

	case CTL_PREFIXPOS:
		if (state->tabprefix)
			draw_line_output(state, state->tabprefix);
		else if (state->curprefix)
			draw_line_output(state, state->curprefix);
		break;

	case CTL_TEXT:
		if (!cp->ctc_value)
			break;
		if (state->align == AL_LEFTFILL0) {
			state->align = AL_LEFTFILL1;
			state->leftfillpos = state->linewidth;
		}
		draw_line_output(state, cp->ctc_value);
		break;

	case CTL_LINESTART:
		if (state->line == 0) {
			if (bg_image) {
				if (bg_image_prev
				 && ctlcmp(bg_image, bg_image_prev) == 0) {
					; /* same as the last time */
				} else if (bg_image->ct_op == CTL_BIMAGE) {
				    image_load(state, bg_image->ctm_fname,
					bg_image->ctm_numcolor,
					bg_image->ctm_ximagesize,
					bg_image->ctm_yimagesize, 1,
					bg_image->ctm_zoomflag, 0);
				} else if (bg_image->ct_op == CTL_BGRAD) {
					back_gradation(state,
						&bg_image->ct_val.ctrl_grad);
				} else if (bg_image->ct_op == CTL_BACK) {
					XSetWindowBackground(display,
						state->target,
						bg_image->ctl_value);
					back_color = bg_image->ctl_value;
					XSetForeground(display, gcback,
						back_color);
					XSetBackground(display, gcfb,
						back_color);
				} else {
					fprintf(stderr, "bogus background\n");
					exit(1);
				}
			}
			XClearWindow(display, state->target);
			bg_image_prev = bg_image;
			bg_image = NULL;
			if (t_fin)
				timebar(state);
		}
		draw_line_start(state);
		break;

	case CTL_LINEEND:
		draw_line_end(state);
		/* reset single-line oriented state */
		state->tabprefix = NULL;
		state->special = 0;
		if (state->align == AL_LEFTFILL1) {
			state->align = AL_LEFTFILL0;
			state->leftfillpos = 0;
		}
		break;

	case CTL_MARK:
		state->have_mark = 1;
		state->mark_ypos = state->ypos;
		break;

	case CTL_SYSTEM:
		process_system(state, cp);
		break;

	case CTL_ICON:
		process_icon(state, cp);
		break;

	default:
		break;
	}
}

/*
 * line management.
 */
static int
set_position(state)
	struct render_state *state;
{
	int x;

	x = 0;
	switch (state->align) {
	case AL_CENTER:
		x = (window_width - state->linewidth)/ 2;
		break;

	case AL_LEFT:
	case AL_LEFTFILL0:
	case AL_LEFTFILL1:
		x = 0;
		break;

	case AL_RIGHT:
		x = window_width - state->linewidth;
		break;
	}

	return x;
}

static void
draw_line_start(state)
	struct render_state *state;
{
	state->lineheight = 0;
	draw_line_itemsize(state, VERT_STEP(char_size));
	state->linewidth = 0;
}

static void
draw_line_itemsize(state, newsize)
	struct render_state *state;
	u_int newsize;
{
	u_int oldsize, upos, lpos, dy;

	oldsize = state->lineheight;
	if (newsize <= oldsize)
		return;
	state->lineheight = newsize;
#if 1
	upos = PM_POSY(newsize);
	if (oldsize == 0) {
		dy = ((newsize + 1) / 2) * 2;
		if (COMPLEX_BGIMAGE) {
			XFillRectangle(display, pixmap, gcnone,
				0, upos, window_width, dy);
			if (fore_color != depth_mask)
				XFillRectangle(display, maskpix, gcnone,
					0, upos, window_width, dy);
		} else {
			XFillRectangle(display, pixmap, gcback,
				0, upos, window_width, dy);
		}
	} else {
		lpos = window_height / 2 + (oldsize + 1) / 2;
		dy = (newsize + 1) / 2 - (oldsize + 1) / 2;
		if (dy == 0)
			return;
		if (COMPLEX_BGIMAGE) {
			XFillRectangle(display, pixmap, gcnone,
				0, upos - dy, window_width, dy);
			XFillRectangle(display, pixmap, gcnone,
				0, lpos, window_width, dy);
			if (fore_color != depth_mask) {
				XFillRectangle(display, maskpix, gcnone,
					0, upos - dy, window_width, dy);
				XFillRectangle(display, maskpix, gcnone,
					0, lpos, window_width, dy);
			}
		} else {
			XFillRectangle(display, pixmap, gcback,
				0, upos, window_width, dy);
			XFillRectangle(display, pixmap, gcback,
				0, lpos, window_width, dy);
		}
	}
#endif
}


static void
draw_line_output(state, data)
	struct render_state *state;
	u_char *data;
{
	draw_line_itemsize(state, VERT_STEP(char_size));
	draw_string(state, data);
}

static void
draw_line_end(state)
	struct render_state *state;
{
	u_int width;
	u_int xpos;

	xpos = set_position(state);

	/* process the special attribute. */
	switch (state->special) {
#if 0
	case SP_SHRINK:
		shrink(data, page, xpos);
		break;
#endif
	case SP_LCUTIN:
		cutin(state, pixmap, maskpix,
			(int)xpos, VERT_POS(state->ypos, char_size), 1);
		break;
	case SP_RCUTIN:
		cutin(state, pixmap, maskpix,
			(int)xpos, VERT_POS(state->ypos, char_size), -1);
		break;
	default:
		break;
	}

	width = (state->linewidth <= window_width - xpos)
			? state->linewidth
			: window_width - xpos;

	if (COMPLEX_BGIMAGE) {
		if (fore_color != depth_mask)
			XCopyArea(display, maskpix, window, gcmask,
				0, PM_POSY(state->lineheight),
				width, state->lineheight,
				xpos, VERT_POS(state->ypos, char_size));
		XCopyArea(display, pixmap, window, gcor,
			0, PM_POSY(state->lineheight), width, state->lineheight,
			xpos, VERT_POS(state->ypos, char_size));
	} else {
		XCopyArea(display, pixmap, window, gcall,
			0, PM_POSY(state->lineheight), width, state->lineheight,
			xpos, VERT_POS(state->ypos, char_size));
	}

	/* cleanup backing store */
	if (COMPLEX_BGIMAGE) {
		XFillRectangle(display, pixmap, gcnone,
			0, window_height / 2 - state->lineheight,
			window_width, state->lineheight * 2);
		if (fore_color != depth_mask) {
			XFillRectangle(display, maskpix, gcnone,
				0, window_height / 2 - state->lineheight,
				window_width, state->lineheight * 2);
		}
	} else {
		XFillRectangle(display, pixmap, gcback,
			0, window_height / 2 - state->lineheight,
			window_width, state->lineheight * 2);
	}

	state->ypos += state->lineheight;
}

static void
cutin(state, pixmap, maskpix, lx, ly, dir)
	struct render_state *state;
	Pixmap maskpix;
	Pixmap pixmap;
	int lx;
	int ly;
	int dir;
{
	u_int step, x;
	int i;
	int sx;
	Window cutinWin;

	sx = (0 < dir) ? 0 : window_width - state->linewidth;
	step = (lx - sx) / 20;
	if (!step)
		step = 1;

	cutinWin = XCreateSimpleWindow(display, state->target,
		sx, ly, state->linewidth, state->lineheight, 0, fore_color, back_color);
	XSetWindowBackgroundPixmap(display, cutinWin, None);
	XMapSubwindows(display, state->target);

	if (COMPLEX_BGIMAGE) {
		XCopyArea(display, maskpix, cutinWin, gcmask,
			0, PM_POSY(state->lineheight),
			state->linewidth, state->lineheight, 0, 0);
		XCopyArea(display, pixmap, cutinWin, gcor,
			0, PM_POSY(state->lineheight),
			state->linewidth, state->lineheight, 0, 0);
	} else {
		XCopyArea(display, pixmap, cutinWin, gcall,
			0, PM_POSY(state->lineheight),
			state->linewidth, state->lineheight, 0, 0);
	}
	XFlush(display);

	x = sx;
	for (i = 0; i < 20; i++) {	/*XXX*/
		XMoveWindow(display, cutinWin, x, ly);
		XFlush(display);
		usleep(CUTIN_DELAY);
		x += step;
	}

	XDestroyWindow(display, cutinWin);
}

#if 0
static void
shrink(data, page)
	u_char *data;
	u_int page;
{
	u_int min_csize = char_size;
	u_int max_csize = window_height / 4;
	u_int csize, i, x;
	u_int step = (max_csize - min_csize) / 3;

	if (!step)
		step = 1;

	if (state->align != AL_CENTER) {
		fprintf(stderr, "align is not center: \n");
		return;
	}

	csize = char_size;
	for (i = max_csize; i > min_csize; i -= step) {
		char_size = i;
		draw_string(state, data);
		x = (window_width - state->linewidth) / 2;
		XCopyArea(display, maskpix, state->target, gc,
			0, 0, state->linewidth, char_size, x, state->ypos);
		XCopyArea(display, pixmap, state->target, gcor,
			0, 0, state->linewidth, char_size, x, state->ypos);
		XFlush(display);
		usleep(SHRINK_DELAY);
		XFillRectangle(display, pixmap, gcall,
			0, 0, window_width, char_size);
		XFillRectangle(display, maskpix, gcall,
			0, 0, window_width, char_size);
		XClearArea(display, state->target, x, state->ypos,
			state->linewidth, char_size, 0);
	}
	char_size = csize;
}
#endif

/*
 * render characters.
 */
static void
draw_string(state, data)
	struct render_state *state;
	u_char *data;
{
	u_char *p;
	u_int kanji = 0;
	u_int code;
	u_int char_len;

	p = data;

	while (*p != '\n' && *p != 0) {
		if (*p == 0x1b && *(p+1) == '$' &&
				(*(p+2) == 'B' || *(p+2) == '@')) {
			kanji = 1;
			p += 3;
			continue;
		}
		if (*p == 0x1b && *(p+1) == '(' &&
				(*(p+2) == 'B' || *(p+2) == 'J')) {
			kanji = 0;
			p += 3;
			continue;
		}

		if (kanji) {
			code = *p * 256 + *(p + 1);
			p += 2;
		} else {
			code = *p;
			p++;
		}

		if (code == ' ') {
			state->linewidth += char_size / 2;
			continue;
		}

#ifdef USE_XDRAWSTRING
		char_len = draw_onechar_x(state, code,
			state->linewidth, PM_POSY(char_size));
#else
# ifdef USE_XDRAWSTRING_ONLY_SMALL
		if (!kanji && char_size < 25) {
			char_len = draw_onechar_x(state, code,
				state->linewidth, PM_POSY(char_size));
		} else
# endif
		{
			if (!kanji)
				code = ctable[code];
			char_len = draw_onechar_vf(state, code,
				state->linewidth, PM_POSY(char_size),
				char_size, char_size);
		}
#endif

		state->linewidth += HORIZ_STEP(char_size, char_len);

		if (state->align == AL_LEFTFILL1
		 && window_width < state->linewidth + char_size) {
			draw_line_end(state);
			draw_line_start(state);
			state->linewidth = state->leftfillpos;
		}
	}
}

#ifdef VFLIB
static u_int
draw_onechar_vf(state, code, x, y, width, height)
	struct render_state *state;
	u_int code;
	u_int x, y, width, height;
{
	struct vfont *vfc;
	static u_long prev_color;
	XImage *xim = NULL;

	vfc = vfc_get(code, width, height, 1);

	if (COMPLEX_BGIMAGE) {
		XPutImage(display, pixmap, gcfore, vfc->image,
			vfc->minx, 0, x, y, vfc->charlen, vfc->height);
		if (fore_color != depth_mask)
			XPutImage(display, maskpix, gcall, vfc->image,
				vfc->minx, 0, x, y, vfc->charlen, vfc->height);
	} else {
#ifdef ANTIALIAS
		if (!(mgp_flag & FL_NOALIAS))
			xim = vfc_image(vfc, fore_color, back_color);
#endif /*ANTIALIAS*/
		if (!xim)
			xim = vfc->image;
		XPutImage(display, pixmap, gcfb, xim,
			vfc->minx, 0, x, y, vfc->charlen, vfc->height);
		if (xim != vfc->image)
			XDestroyImage(xim);
	}
	return vfc->charlen;
}
#endif /*VFLIB*/

static u_int
draw_onechar_x(state, code, x, y)
	struct render_state *state;
	u_int code;
	u_int x, y;
{
	u_char fontstring[BUFSIZ];
	static u_int last_char_size = 0;
	static u_char last_xfont[BUFSIZ] = "";
	u_int i;
	u_int char_len;
	u_int char_height;
	static XFontStruct *xfontstruct;
	static int lastkanjimode = -1;
	int kanjimode;

	kanjimode = (code & 0xff00) ? 1 : 0;

	if (strcmp(state->xfont, last_xfont) || char_size != last_char_size
	 || kanjimode != lastkanjimode) {
		/* load new font */
		if (xfontstruct)
			XFreeFont(display, xfontstruct);

#if 1
		/*
		 * Look for the best font possible.
		 * 1. Check for a font that is smaller than the required one.
		 *    By using smaller font, we won't make the screen garbled.
		 * 2. If 1. is impossible, look for slightly larger font than
		 *    the required one.
		 */
#define CHKFONTSIZE(x)	((char_size < x) ? x : (char_size - x))
#define	FONTSEARCH	20
#else
		/* The original one. */
#define CHKFONTSIZE(x)	((int)char_size - x)
#define	FONTSEARCH	20
#endif
		for (i = 0; i < FONTSEARCH; i++) {
			if (CHKFONTSIZE(i) < 0)
				continue;
			if (kanjimode) {
				sprintf(fontstring, KANJIFONT_FORMAT,
					CHKFONTSIZE(i));
			} else {
				sprintf(fontstring, FONT_FORMAT, state->xfont,
					CHKFONTSIZE(i));
			}
			if (xfontstruct = XLoadQueryFont(display, fontstring))
				break;
		}
#undef CHKFONTSIZE
#undef FONTSEARCH
		if (xfontstruct == NULL) {
			if (kanjimode) {
				sprintf(fontstring, KANJIFONT_FORMAT,
					char_size);
			} else {
				sprintf(fontstring, FONT_FORMAT, state->xfont,
					char_size);
			}
			fprintf(stderr, "can't load font %s "
				"(nor font in similar size)\n", fontstring);
			exit(-1);
		}
		XSetFont(display, gcfore, xfontstruct->fid);
		XSetFont(display, gcfb, xfontstruct->fid);
		XSetFont(display, gcall, xfontstruct->fid);

		last_char_size = char_size;
		strcpy(last_xfont, state->xfont);
	}
	lastkanjimode = kanjimode;

#define YCONTROL	(xfontstruct->descent / 2)
	draw_line_itemsize(state,
		(char_size > xfontstruct->ascent)
			? char_size + xfontstruct->descent + 4
			: xfontstruct->ascent + xfontstruct->descent + 4);
	if (kanjimode) {
		XChar2b kch[2];

		kch[0].byte1 = (code >> 8) & 0xff;
		kch[0].byte2 = code & 0xff;

		char_len = XTextWidth16(xfontstruct, kch, 1);

		if (COMPLEX_BGIMAGE) {
			XDrawString16(display, pixmap, gcfore,
				x, y + char_size - YCONTROL, kch, 1);
			XDrawString16(display, maskpix, gcall,
				x, y + char_size - YCONTROL, kch, 1);
		} else {
			XDrawString16(display, pixmap, gcfb,
				x, y + char_size - YCONTROL, kch, 1);
		}
	} else {
		u_char ch[2];
		ch[0] = code & 0xff;

		char_len = XTextWidth(xfontstruct, ch, 1);

		if (COMPLEX_BGIMAGE) {
			XDrawString(display, pixmap, gcfore,
				x, y + char_size - YCONTROL, ch, 1);
			XDrawString(display, maskpix, gcall,
				x, y + char_size - YCONTROL, ch, 1);
		} else {
			XDrawString(display, pixmap, gcfb,
				x, y + char_size - YCONTROL, ch, 1);
		}
	}
#undef YCONTROL

	return char_len;
}

/*
 * render misc items.
 */
static void
back_gradation(state, cg0)
	struct render_state *state;
	struct ctrl_grad *cg0;
{
	struct ctrl_grad cg1;
	struct ctrl_grad *cg;
	int srcwidth, srcheight;
	int dstwidth, dstheight;
	int dir, numcolor;
	int xzoomrate, yzoomrate;
	int hquality, vquality;

	Image *myimage, *image;
	Pixmap mypixmap;
	XImageInfo *ximageinfo;
	Visual *visual;
	byte *pic;
	int verbose = mgp_flag & FL_VERBOSE;
	static Cursor curs;
	u_int depth = DEFAULT_GRADDEPTH;

	/* okay, please wait for a while... */
	if (!curs)
		curs = XCreateFontCursor(display, XC_watch);
	XDefineCursor(display, state->target, curs);
	XFlush(display);

	/* just for safety */
	memcpy(&cg1, cg0, sizeof(struct ctrl_grad));
	cg = &cg1;

	/* grab parameters */
	dir = cg->ct_direction;
	numcolor = cg->ct_numcolor;
	hquality = b_quality;
	vquality = b_quality;

	/*
	 * XXX zoomflag is too complex to understand.
	 */
	if (!cg->ct_zoomflag) {
		int t;
		int i;

		dstwidth = window_width * cg->ct_width / 100;
		dstheight = window_height * cg->ct_height / 100;
		srcwidth = dstwidth;
		srcheight = dstheight;

		/*
		 * apply quality factor if srcwidth/height are large enough.
		 */
#define TOOSMALLFACTOR 8
		t = srcwidth;
		for (i = 100; hquality < i; i--) {
			t = srcwidth * i / 100;
			if (t < cg->ct_g_colors * TOOSMALLFACTOR)
				break;
		}
		srcwidth = t;

		t = srcheight;
		for (i = 100; vquality < i; i--) {
			t = srcheight * i / 100;
			if (t < cg->ct_g_colors * TOOSMALLFACTOR)
				break;
		}
		srcheight = t;
#undef TOOSMALLFACTOR
	} else {
		int t;
		int i;

		dstwidth = window_width;
		dstheight = window_height;
		srcwidth = window_width * cg->ct_width / 100;
		srcheight = window_height * cg->ct_height / 100;

		/*
		 * we don't apply quality factor here, since srcwidth/height
		 * is already smaller than dstwidth/height.
		 */
	}

#if 0
	if (srcwidth * hquality / 100 < cg->ct_g_colors * TOOSMALLFACTOR
	 || srcheight * vquality / 100 < cg->ct_g_colors * TOOSMALLFACTOR) {
		srcwidth = srcwidth * hquality / 100;
		srcheight = srcheight * vquality / 100;
	}
#endif

	xzoomrate = 100 * dstwidth / srcwidth;
	yzoomrate = 100 * dstheight / srcheight;

	/* performace enhance hack for special case */
	if (dir % 90 == 0) {
		int *p;
		int *q;
		int primes[] = { 2, 3, 5, 7, 11, 13, 0 };
		int i;

		/*
		 * 0 or 180: reduce width
		 * 90 or 270: reduce height
		 */
		p = (dir % 180 == 0) ? &srcwidth : &srcheight;
		q = (dir % 180 == 0) ? &xzoomrate : &yzoomrate;

		for (i = 0; primes[i]; i++) {
			while (*p % primes[i] == 0) {
				*p = *p / primes[i];
				*q = *q * primes[i];
			}
		}
	}

	if (verbose) {
		fprintf(stderr, "raw: %d,%d qu: %d,%d "
			"dst: %d,%d src: %d,%d zoom: %d,%d\n",
			cg->ct_width, cg->ct_height,
			hquality, vquality,
			dstwidth, dstheight, srcwidth, srcheight,
			xzoomrate, yzoomrate);
	}

	screen = DefaultScreen(display);

	/* make gradation image */
	pic = draw_gradation(srcwidth, srcheight, cg);
	myimage = make_XImage(pic, srcwidth, srcheight);

	if (numcolor < 64)
		myimage = reduce(myimage, numcolor, verbose);

#ifdef DEBUG
	fprintf(stderr, "Draw:zoomrate:(%d,%d)->(%d,%d)\n",
		cg->ct_xzoom,cg->ct_yzoom,xzoomrate,yzoomrate);
	fprintf(stderr, "zoom mode %d: (%d, %d)->(%d, %d)[%d]\n",
		zoomflag, width, height, window_width, window_height,
		b_quality);
#endif

	if (xzoomrate != 100 || yzoomrate != 100) {
		image = myimage;
		myimage = zoom(image, xzoomrate, yzoomrate, verbose);
		freeImage(image);
	}

	/* check visual class */
	bestVisual(display, screen, myimage, &visual, &depth);
	if (visual != DefaultVisual(display, screen)) {
		if (verbose) {
			fprintf(stderr, "  Using %s visual\n",
				nameOfVisualClass(visual->class));
		}
	}

	ximageinfo = imageToXImage(display, screen, visual, depth, myimage,
		0, 0, verbose);
	if (!ximageinfo) {
		fprintf(stderr, "Cannot convert Image to XImage\n");
		exit(1);
	}

	mypixmap = ximageToPixmap(display, DefaultRootWindow(display),
		ximageinfo);
	if (mypixmap == None && verbose) {
		fprintf(stderr, "Cannot create image in server\n");
		exit(1);
	}

	/* draw background */
	XSetWindowBackgroundPixmap(display, state->target, mypixmap);

	/* finish */
	XFreePixmap(display, mypixmap);
	freeXImage(myimage, ximageinfo);
	freeImage(myimage);
	XUndefineCursor(display, state->target); XFlush(display);
}

static void
image_load(state, filename, numcolor, ximagesize, yimagesize, backflag, zoomflag, centerflag)
	struct render_state *state;
	char *filename;
	int numcolor;
	int ximagesize;
	int yimagesize;
	int backflag;
	int zoomflag;
	int centerflag;
{
	Image *image, *myimage;
	Pixmap mypixmap;
	XImageInfo *ximageinfo;
	Visual *visual;
	u_int depth, image_posx;
	int width, height;
	int xzoomrate, yzoomrate;
	int	verbose = mgp_flag & FL_VERBOSE;
	static Cursor curs;
	static Pixmap backpix;
	static char backfile[MAXPATHLEN];
	static int backzoom, backnumcolor, backx, backy;

	if (pause_flag)
		return;

	if (!curs)
		curs = XCreateFontCursor(display, XC_watch);
	XDefineCursor(display, state->target, curs); XFlush(display);

	if ((myimage = loadImage(filename, verbose)) == NULL)
		exit(-1);	/* fail to load image data */
	width = myimage->width;
	height = myimage->height;

	if (myimage->depth == 1) {
		XColor xc;
	        Colormap cmap;

	        cmap = DefaultColormap(display, screen);
		xc.flags = DoRed | DoGreen | DoBlue;
		xc.pixel = fore_color;
		XQueryColor(display, cmap, &xc);
		*(myimage->rgb.red + 1) = xc.red;
		*(myimage->rgb.green + 1) = xc.green;
		*(myimage->rgb.blue + 1) = xc.blue;
		xc.pixel = back_color;
		XQueryColor(display, cmap, &xc);
		*myimage->rgb.red = xc.red;
		*myimage->rgb.green = xc.green;
		*myimage->rgb.blue = xc.blue;
	}

	if (numcolor)
		myimage = reduce(myimage, numcolor, verbose);

	if (zoomflag == 2) {
		/*
		 * auto resize according to physical and desired screen size.
		 * allow 5% error for '-o' option.
		 */
		if (ximagesize == 0 || ximagesize == window_width)
			ximagesize = 100;
		else
			ximagesize = window_width * 100 / ximagesize;
		if (yimagesize == 0 || yimagesize == window_height)
			yimagesize = 100;
		else
			yimagesize = window_height * 100 / yimagesize;
		if (ximagesize > 95 && ximagesize < 105 &&
		    yimagesize > 95 && yimagesize < 105)
			ximagesize = yimagesize = 0;
	}
	if (ximagesize != 0 && yimagesize != 0) {
		image = myimage;
		if (!zoomflag) {
			xzoomrate = window_width * ximagesize / width;
			yzoomrate = window_height * yimagesize / height;
		} else {
			xzoomrate = ximagesize;
			yzoomrate = yimagesize;
		}
		if (xzoomrate != 100 || yzoomrate != 100) {
			myimage = zoom(image, xzoomrate, yzoomrate, verbose);
			freeImage(image);
		}
	} else
		xzoomrate = yzoomrate = 100;

	bestVisual(display, screen, myimage, &visual, &depth);
	if (verbose && (visual != DefaultVisual(display, screen)))
		printf("  Using %s visual\n", nameOfVisualClass(visual->class));

	if (! (ximageinfo= imageToXImage(display, screen, visual, depth, myimage,
		0, 0, verbose))) {
			fprintf(stderr, "Cannot convert Image to XImage\n");
			exit(1);
	}

	if (((mypixmap= ximageToPixmap(display,
		RootWindow(display, screen), ximageinfo)) == None) && verbose) {
			fprintf(stderr, "Cannot create image in server\n");
			exit(1);
	}

	if (backflag) {
		XSetWindowBackgroundPixmap(display, state->target, mypixmap);
		strcpy(backfile, filename);
		backnumcolor = numcolor;
		backzoom = zoomflag;
		backx = ximagesize;
		backy = yimagesize;
		goto end;
	}

	draw_line_itemsize(state, height * yzoomrate / 100);
	if (centerflag){
		image_posx = char_size / 2 - (width * xzoomrate / 100) /2;
	} else
		image_posx = 0;

	XCopyArea(display, mypixmap, pixmap, gcall, 0, 0,
		width * xzoomrate / 100, height * yzoomrate / 100,
		state->linewidth + image_posx,
		PM_POSY(height * yzoomrate / 100));
	state->linewidth += (width * xzoomrate / 100);

end:
	XFreePixmap(display, mypixmap);
	freeXImage(myimage, ximageinfo);
	freeImage(myimage);
	XUndefineCursor(display, state->target); XFlush(display);
}

void
timebar(state)
	struct render_state *state;
{
	int pos, n, p, barlen;
	GC *pgc;

	if (t_start == 0 || tbar_mode == 0)
		return;

	pos = (window_width - 2) * (state->page - 1) / (maxpage - 1);
	p = (time(NULL) - t_start) * 100;
	barlen = window_width - window_width * p / t_fin / 6000;

	if (window_width * 50 / 100 < barlen)
		pgc = &gcgreen;
	else if (window_width * 30 / 100 < barlen)
		pgc = &gcyellow;
	else
		pgc = &gcred;
	if (barlen > 0) {
		XFillRectangle(display, state->target, gcback,
			0, window_height - 2, window_width, window_height);
		XFillRectangle(display, state->target, *pgc,
			window_width - barlen, window_height - 1,
			window_width, window_height);
		XFillRectangle(display, state->target, *pgc,
			pos, window_height - 5,
			2, 5);
	} else if (barlen < 0) {
		barlen = - barlen;
		n = p / t_fin / 6000;
		if (n > window_height - 1)
			n = window_height - 1;
		if (n)
			XFillRectangle(display, state->target, gcred,
				0, window_height - n,
				barlen, window_height);
		XFillRectangle(display, state->target, gcback,
			0, window_height - n - 2,
			window_width, window_height - n);
		XFillRectangle(display, state->target, gcred,
			0, window_height - n - 1,
			barlen % window_width, window_height - n);
		XFillRectangle(display, state->target, gcred,
			pos, window_height - n - 1 - 4,
			2, 5);
	}
}

static void
process_icon(state, cp)
	struct render_state *state;
	struct ctrl *cp;
{
	static GC gcicon;
	u_int i, icon_type, icon_size, icon_x, icon_y;
	u_long tmp_color;
	static struct ctl_words icon_words[] = {
		{ 1, "box", 3 },
		{ 2, "arc", 3 },
		{ 0, NULL, 0 }
	};

	if (!gcicon)
		gcicon = XCreateGC(display, state->target, 0, 0);

	for (i = 0; icon_words[i].ctl_strlen != 0; i++) {
		if (!strncasecmp(cp->ctic_value, icon_words[i].ctl_string,
			strlen(cp->ctic_value))) {
				break;
		}
	}

	icon_type = icon_words[i].ctl_type; /* may be 0 */
	icon_size = char_size * cp->ctic_size / 100;

	switch(icon_type){
	case 0:
		/* this is image */
		icon_x = icon_size * 100 / window_width;
		icon_y = icon_size * 100 / window_height;
		if (icon_x == 0) icon_x = 1;
		if (icon_y == 0) icon_y = 1;
		tmp_color = fore_color;
		fore_color = cp->ctic_color;
		image_load(state, cp->ctic_value, 0, icon_x, icon_y, 0, 0, 1);
		fore_color = tmp_color;
		break;

	case 1:
		/* this is box */
		XSetForeground(display, gcicon, cp->ctic_color);
		XFillRectangle(display, pixmap, gcicon,
			state->linewidth + char_size /2 - icon_size /2,
			PM_POSY(icon_size), icon_size, icon_size);
		state->linewidth += char_size;
		break;

	case 2:
		/* this is arc */
		XSetForeground(display, gcicon, cp->ctic_color);
		XFillArc(display, pixmap, gcicon,
			state->linewidth + char_size /2 - icon_size /2 ,
			PM_POSY(icon_size), icon_size, icon_size, 0, 360 * 64);
		state->linewidth += char_size;
		break;

	default:
		break;
	}

	cp = NULL;
}

static void
draw_bar(state, cp)
	struct render_state *state;
	struct ctrl *cp;
{
	u_int width, swidth, st, len;
	Colormap cmap;
	Visual *visual;
	XColor col, scol;
	static GC gcbar, gcsbar;
	static u_long prevcolor = -1;

	if (!gcbar) {
		gcbar = XCreateGC(display, state->target, 0, 0);
		XSetFunction(display, gcbar, GXcopy);
		gcsbar = XCreateGC(display, state->target, 0, 0);
		XSetFunction(display, gcsbar, GXcopy);
	}
	col.pixel = cp->ctb_color;
	if (col.pixel == -1)
		col.pixel = fore_color;
	if (col.pixel != prevcolor) {
		prevcolor = col.pixel;
		visual = DefaultVisual(display, screen);
		cmap = DefaultColormap(display, screen);
		col.flags = DoRed|DoGreen|DoBlue;
		XQueryColor(display, cmap, &col);
		scol.red   = col.red   / 2;
		scol.green = col.green / 2;
		scol.blue  = col.blue  / 2;
		if (!XAllocColor(display, cmap, &scol))
			scol.pixel = col.pixel;
		XSetForeground(display, gcbar, col.pixel);
		XSetForeground(display, gcsbar, scol.pixel);
	}
	width = cp->ctb_width * window_height / 1000;
	swidth = width / 2;
	width -= swidth;
	st = cp->ctb_start * window_width / 100;
	len = cp->ctb_length * window_width / 100;
	XFillRectangle(display, state->target, gcbar, st, state->ypos, len, width);
	XFillRectangle(display, state->target, gcsbar, st, state->ypos + width, len, swidth);

	state->ypos += width + swidth + VERT_GAP(char_size) / 2;
}

static void
process_system(state, cp)
	struct render_state *state;
	struct ctrl *cp;
{
	pid_t pid;
	int status;
	int i;
	char **argv;
	char buf[BUFSIZ];
	int x, y;
	int rootxsiz, rootysiz;

    {
	XWindowAttributes wa;
	Window junkwin;
	int junk;

	XGetWindowAttributes(display, window, &wa);
	XTranslateCoordinates(display, window, wa.root,
		-wa.border_width, -wa.border_width, &x, &y, &junkwin);
	XGetGeometry(display, wa.root, &junkwin, &junk, &junk,
		&rootxsiz, &rootysiz, &junk, &junk);
     }

	if (checkchild(cp) != (pid_t)-1)
		return;	/*already running*/

	/*
	 * edit argument.
	 * if we have X11 geometry string 
	 */
	argv = (char **)cp->cta_argv;
	for (i = 0; i < cp->cta_argc; i++) {
		if (*(argv[i]) == '%')
			break;
	}
	if (i < cp->cta_argc) {
		char *p;
		char *q;
		int xsiz, ysiz;
		int xloc, yloc;
		char xdir, ydir;

		argv = (char **)malloc((cp->cta_argc + 1) * sizeof(char *));
		memcpy(argv, cp->cta_argv, (cp->cta_argc + 1) * sizeof(char *));
		p = argv[i];
		p++;	/*drop percent char*/
		q = buf;
		*q = '\0';

		xsiz = ysiz = -1;
		if (isdigit(*p)) {
			xsiz = atoi(p);
			while (isdigit(*p))
				p++;
			if (*p++ != 'x')
				goto fail;
			ysiz = atoi(p);
			while (isdigit(*p))
				p++;

			sprintf(q, "%dx%d", xsiz * window_width / 100,
				ysiz * window_height / 100);
			q += strlen(q);
		}
		xloc = yloc = 0;
		if (*p == '-' || *p == '+') {
			xdir = *p++;
			xloc = atoi(p);
			while (isdigit(*p))
				p++;
			if (*p != '-' && *p != '+')
				goto fail;
			ydir = *p++;
			yloc = atoi(p);
			while (isdigit(*p))
				p++;
			if (*p != '\0')
				goto fail;

			xloc = xloc * window_width / 100;
			yloc = yloc * window_height / 100;
			if (xdir == '+')
				xloc += x;
			else
				xloc = rootxsiz - x + window_width - xloc;
			if (ydir == '+')
				yloc += y;
			else
				yloc = rootysiz - y + window_height - yloc;
			sprintf(q, "+%d+%d", xloc, yloc);
		}

		if (mgp_flag & FL_VERBOSE) {
			fprintf(stderr, "relative geometry: "
				"%s (presentation %dx%d+%d+%d)\n",
				argv[i], window_width, window_height, x, y);
			fprintf(stderr, "\t-> %s\n", buf);
		}
		argv[i] = buf;

		if (0) {
fail:
			if (mgp_flag & FL_VERBOSE) {
				fprintf(stderr,
					"relative geometry: %s failed\n",
					argv[i]);
			}
		}
	}
	pid = fork();
	if (pid < 0) {
		perror("fork");
		exit(1);
	} else if (pid == 0)
		execvp(argv[0], argv);

	if (!cp->cta_flag)	/*will be purged at the end of page*/
		regchild(pid, cp, state->page);
	else
		regchild(pid, cp, cp->cta_flag);
}
