/*
 * Copyright (C) 1997 WIDE Project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by WIDE Project and
 *	its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/*
 * $Id: mgp.c,v 1.38.2.33 1997/12/26 04:44:26 kazu Exp $
 */

static char *mgp_version = "1.01a (971226/itojun branch)";

#include "mgp.h"

Window plwin[MAXPAGE];
Pixmap maskpix;
XFontStruct *plfs;
XFontStruct *plkfs;

u_int pause_flag;

u_int pg_mode;
u_int pl_mode;
u_int pl_tmode;
u_int pl_nx, pl_ny, pl_y;
u_long pl_fh, pl_fw;
time_t t_start;
u_int t_fin;
u_int tbar_mode;

int rakugaki = 0;
int rakugaki_x = -1;
int rakugaki_y = -1;

u_long depth_mask;

char *back_clname = DEFAULT_BACK;

#ifdef VFLIB
char *vfcap_name = VFCAP;
#endif

static void mgp_usage __P((char *));
static void mgp_show_version __P((char *));
#ifdef VFLIB
static void vf_init __P((void));
#endif
static void beep __P((void));
static void main_loop __P((u_int));
static void set_geometry __P((u_char *));
static void rakugaki_update __P((XEvent *));
static RETSIGTYPE waitkids __P((int));

/*image*/
extern char *expandPath __P((char *));

static void
cleanup()
{
	signal(SIGTERM, SIG_IGN);
	killpg(0, SIGTERM);
}

int
main(argc, argv)
	int argc;
	char **argv;
{
	u_int opt;
	extern char *optarg;
	extern int optind;
	u_char *progname;
	u_int start_page = 1;
	char buf[BUFSIZ], *p;
	extern char *Paths[];
	extern int NumPaths;
#ifdef VFLIB
	char *vfont_name = VFONT;
#endif

	progname = argv[0];

	while ((opt = getopt(argc, argv, "ABdvVob:c:g:f:hlGp:qt:Q:P")) != -1) {
		switch (opt) {
		case 'A':
			mgp_flag |= FL_NOALIAS;
			break;

		case 'B':
			mgp_flag |= FL_BIMAGE;
			break;

		case 'd':
			mgp_flag |= FL_DEMO;
			break;

		case 'V':
			mgp_flag |= FL_VERBOSE;
			break;

		case 'l':
			mgp_flag |= FL_OUTLINE;
			break;

		case 'G':
			pg_mode = 1;
			break;

		case 'o':
			mgp_flag |= FL_OVER;
			break;

#ifdef VFLIB
		case 'c':
			vfcap_name = optarg;
			break;

		case 'f':
			vfont_name = optarg;
			break;
#endif /*VFLIB*/

		case 'b':
			back_clname = optarg;
			break;

		case 'p':
			start_page = atoi(optarg);
			break;

		case 'g':
			set_geometry(optarg);
			break;

		case 'q':
			mgp_flag |= FL_NOBEEP;
			break;

		case 't':
			t_fin = atoi(optarg);
			tbar_mode = 1;
			break;

		case 'Q':
			b_quality = atoi(optarg);
			quality_flag = 1;
			break;

		case 'P':
			parse_debug++;
			break;

		case 'v':
			mgp_show_version(progname);
			/*NOTREACHED*/

		case 'h':
		default:
			mgp_usage(progname);
			/*NOTREACHED*/
		}
	}

	argc -= optind;
	argv += optind;

	if (argc != 1) {
		mgp_usage(progname);
		/*NOTREACHED*/
	}
	mgp_fname = argv[0];

	init_win1();
	strncpy(buf, mgp_fname, sizeof(buf));
	if (p = rindex(buf, '/')) {
		*p = '\0';
        	Paths[NumPaths++]= expandPath(buf);
	}
	loadPathsAndExts();
	load_file(mgp_fname);
	if (parse_error)
		exit(-1);
	init_win2();
#ifdef VFLIB
	vfc_setfont(vfont_name);
#endif

	signal(SIGCHLD, waitkids);
	atexit(cleanup);

	if (mgp_flag & FL_DEMO) {
		struct render_state state;
		memset(&state, 0, sizeof(struct render_state));
		state.target = window;	/*XXX*/
		while (start_page <= maxpage) {
			state_goto(&state, start_page);
			draw_page(&state);
			start_page++;
		}
	} else
		main_loop(start_page);

	exit(0);
}

static void
mgp_show_version(name)
	char *name;
{
	char *p;

	if (p = strrchr(name, '/'))
		p ++;
	else
		p = name;
	fprintf(stderr, "%s version %s\n", p, mgp_version);
	exit(0);
}

static void
mgp_usage(name)
	char *name;
{
	fprintf(stderr, "Usage: %s [opts] mgpfile\n", name);
	fprintf(stderr, "\t-B: Ignore background image\n");
	fprintf(stderr, "\t-d: Demo page is prepended\n");
	fprintf(stderr, "\t-V: Verbose\n");
	fprintf(stderr, "\t-l: Do not use outline font\n");
	fprintf(stderr, "\t-o: Controlled by the window manager\n");
	fprintf(stderr, "\t-G: Page guide is on\n");
#ifdef VFLIB
	fprintf(stderr, "\t-v <vfcap>: Specify vfontcap file\n");
	fprintf(stderr, "\t-f <font>: Specify fontname\n");
#endif /*VFLIB*/
	fprintf(stderr, "\t-b <color>: Specify background color\n");
	fprintf(stderr, "\t-p <page>: Start at the specified page\n");
	fprintf(stderr, "\t-g <geometry>: Set window geometry\n");
	fprintf(stderr, "\t-Q <quality>: Set background image quality(0-100)\n");
	fprintf(stderr, "\t-q Do not beep on errors\n");
	fprintf(stderr, "\t-t <timeslot>: Enable presentation timer\n");
	exit(0);
}

static void
beep()
{
	if (!(mgp_flag & FL_NOBEEP))
		XBell(display, 0);
}

static void
main_loop(start_page)
	u_int start_page;
{
	XEvent e;
	KeySym key;
	u_int i;
	u_int number = 0;
	u_int shift = 0;
	u_int control = 0;
	static struct render_state state;
	static Cursor pen_curs;

	memset(&state, 0, sizeof(struct render_state));
	state.target = window;	/*XXX*/
	if (!pen_curs)
		pen_curs = XCreateFontCursor(display, XC_pencil);
	state_goto(&state, start_page);
	draw_page(&state);

	while (1) {
		if (!t_start && 1 < state.page)
			t_start = time(NULL);
		if (rakugaki)
			XDefineCursor(display, window, pen_curs);
		else
			XUndefineCursor(display, window);
		XNextEvent(display, &e);
		if (t_fin)
			timebar(&state);
		switch (e.type) {
		case EnterNotify:
			for (i = 1; i <= maxpage; i++) {
				if (e.xany.window == plwin[i]) {
					XClearWindow(display, plwin[i]);
					pl_pdraw(i, gc_plrev);
					pl_title(i);
				}
			}
			XFlush(display);
			break;

		case LeaveNotify:
			for (i = 1; i <= maxpage; i++) {
				if (e.xany.window == plwin[i]) {
					XClearWindow(display, plwin[i]);
					pl_pdraw(i, gc_pl);
					pl_title(0);
				}
			}
			XFlush(display);
			break;

		case ButtonPress:
			if (rakugaki && e.xany.window == window) {
				rakugaki_update(&e);
				break;
			}

			for (i = 1; i <= maxpage; i++) {
				if (e.xany.window == plwin[i]) {
					state_goto(&state, i);
					pause_flag = 0;
					pl_off();
					draw_page(&state);
					pl_tmode = 0;
					break;
				}
			}
			if (e.xany.window == window) {
				if (e.xbutton.button == 1) {
					if (shift)
						pause_flag = 0;
					if (!pause_flag) {
						if (state.page < maxpage) {
						    state_next(&state);
						    draw_page(&state);
						} else
						    beep();
					} else
						draw_page(&state);
				} else {
					if (e.xbutton.button == 3) {
						pause_flag = 0;
						if (state.page > 1) {
						    state_goto(&state, state.page - 1);
						    draw_page(&state);
						} else
						    beep();
					}
				}
			}
			break;

		case ButtonRelease:
			if (rakugaki && e.xany.window == window)
				rakugaki_update(&e);
			break;

		case MotionNotify:
			if (rakugaki && e.xany.window == window)
				rakugaki_update(&e);
			break;

		case KeyPress:
			key = XLookupKeysym((XKeyEvent *)&e, 0);

			switch (key) {
			case XK_q:
			case XK_Escape:
				exit(0);
				break;

			case XK_f:
			case XK_j:
			case XK_n:
			case XK_Down:
			case XK_space:
				if (!pause_flag) {
					struct render_state tstate;
					tstate = state;

					state_next(&tstate);
					if (memcmp(&state, &tstate,
						sizeof(state)) == 0) {
						beep();
					} else if (number == 0
						&& tstate.page <= maxpage) {
						state = tstate;
						draw_page(&state);
					} else if (state.page + number
							<= maxpage) {
						state_goto(&state, state.page + number);
						draw_page(&state);
					} else
						beep();
				} else
					draw_page(&state);
				number = 0;
				break;

#ifdef VFLIB
#ifdef ANTIALIAS
			case XK_a:
				if (shift) {
					mgp_flag ^= FL_NOALIAS;
					state_goto(&state, state.page);
					draw_page(&state);
				}
				number = 0;
				break;
#endif /*ANTIALIAS*/
#endif /*VFLIB*/

			case XK_b:
			case XK_k:
			case XK_p:
			case XK_Up:
			case XK_BackSpace:
			case XK_Delete:
				pause_flag = 0;
				if (number == 0) number = 1;
				if (state.page - number >= 1) {
					state_goto(&state, state.page - number);
					draw_page(&state);
				} else
					beep();
				number = 0;
				break;

			case XK_x:
			    {
				rakugaki = 1 - rakugaki;
				rakugaki_x = rakugaki_y = -1;

				XSelectInput(display, window,
					rakugaki ? EVENT_RAKUGAKI
						 : EVENT_DEFAULT);
				if (rakugaki) {
					XDefineCursor(display, window,
						pen_curs);
				} else
					XUndefineCursor(display, window);
				XFlush(display);
				break;
			    }

			case XK_t:
				if (tbar_mode)
					tbar_mode = 0;
				else {
					if (t_fin)
						tbar_mode = 1;
				}
				break;

			case XK_Control_L:
			case XK_Control_R:
				pl_on();
				control = 1;
				number = 0;
				break;

			case XK_Shift_L:
			case XK_Shift_R:
				shift = 1;
				number = 0;
				break;

			case XK_l:	/* used to be control-L */
				state_goto(&state, state.page);
				draw_page(&state);
				number = 0;
				break;

			case XK_g:
				if (shift) {
					pg_mode = 1 - pg_mode;
					if (pg_mode)
						pg_on(state.page, 1);
					else
						pg_off();
				} else {
					if (number <= maxpage && number > 0) {
						state_goto(&state, number);
						draw_page(&state);
					} else
						beep();
				}
				number = 0;
				break;

			case XK_0:
			case XK_1:
			case XK_2:
			case XK_3:
			case XK_4:
			case XK_5:
			case XK_6:
			case XK_7:
			case XK_8:
			case XK_9:
				number = number * 10 + key - XK_0;
				break;

			default:
				number = 0;
				break;
			}

			break;

		case KeyRelease:
			key = XLookupKeysym((XKeyEvent *)&e, 0);
			switch (key) {
			case XK_Control_L:
			case XK_Control_R:
				pl_off();
				control = 0;
				break;
			case XK_Shift_L:
			case XK_Shift_R:
				shift = 0;
				break;
			}
			break;

		case ConfigureNotify:
			pl_config();
			pause_flag = 0;
			window_width = e.xconfigure.width;
			window_height = e.xconfigure.height;
			state_goto(&state, state.page);
			draw_page(&state);
			break;
		}

		if (state.phase == P_END) {
			if (state.page < maxpage)
				state_goto(&state, state.page + 1);
		}
	}
}

static void
set_geometry(buf)
	u_char *buf;
{
	u_char *p, *p2, tmp[10];

	p = buf;
	strncpy(tmp, p, strlen(p) - strlen(strchr(p, 'x')));
	window_width = atoi(tmp);

	p = strchr(p, 'x') +1;
	p2 = tmp;
	while (*p != '\0' && *p != '+' && *p != '-')
		*(p2++) = *(p++);
	*p2 = 0;
	window_height = atoi(tmp);
	if (*p == '\0')
		return;

	p2 = tmp;
	*(p2++) = *(p++);
	while (*p != '\0' && *p != '+' && *p != '-')
		*(p2++) = *(p++);
	*p2 = 0;

	window_x = atoi(tmp);
	window_y = atoi(p);
}

/* XXX should think about color (= gc) to work with */
static void
rakugaki_update(e)
	XEvent *e;
{
	int x, y;

	if (e->type == MotionNotify) {
		XMotionEvent *em;
		em = (XMotionEvent *)e;
		x = em->x; y = em->y;
	} else if (e->type == ButtonPress) {
		XButtonPressedEvent *eb;
		eb = (XButtonPressedEvent *)e;
		x = eb->x; y = eb->y;
	} else {
		rakugaki_x = rakugaki_y = -1;
		return;
	}

	if (rakugaki_x < 0 || rakugaki_y < 0)
		XDrawRectangle(display, window, gcpen, x, y, 2, 2);
	else {
		XDrawLine(display, window, gcpen,
			x, y,
			rakugaki_x, rakugaki_y);
		XDrawLine(display, window, gcpen,
			x + 1, y,
			rakugaki_x + 1, rakugaki_y);
		XDrawLine(display, window, gcpen,
			x, y + 1,
			rakugaki_x, rakugaki_y + 1);
		XDrawLine(display, window, gcpen,
			x + 1, y + 1,
			rakugaki_x + 1, rakugaki_y + 1);
	}
	rakugaki_x = x;
	rakugaki_y = y;
}

static struct {
	pid_t pid;
	void *key;
	int flag;
} childtab[64];
static int childidx = 0;

pid_t
checkchild(key)
	void *key;
{
	int i;

	for (i = 0; i < childidx; i++) {
		if (childtab[i].pid == (pid_t)-1)
			continue;
		if (childtab[i].key == key)
			return childtab[i].pid;
	}
	return (pid_t)-1;
}

void
regchild(pid, key, flag)
	pid_t pid;
	void *key;
	int flag;
{
	int i;
	for (i = 0; i < childidx; i++) {
		if (childtab[i].pid == (pid_t)-1) {
			childtab[i].pid = pid;
			childtab[i].key = key;
			childtab[i].flag = flag;
			return;
		}
	}
	childtab[childidx].pid = pid;
	childtab[childidx].key = key;
	childtab[childidx].flag = flag;
	childidx++;
}

void
purgechild(flag)
	int flag;
{
	int i;

	for (i = 0; i < childidx; i++) {
		if (childtab[i].pid == (pid_t)-1)
			continue;
		if (childtab[i].flag == flag)
			kill(childtab[i].pid, SIGTERM);
	}
}

static RETSIGTYPE
waitkids(sig)
	int sig;
{
	int status;
	int i;
	pid_t pid;

	if (sig != SIGCHLD) {
		fprintf(stderr, "signal different from expected: %d\n", sig);
		exit(1);
	}
	pid = waitpid(-1, &status, WNOHANG);
	if (pid < (pid_t)0)
		return;
	for (i = 0; i < childidx; i++) {
		if (childtab[i].pid == pid)
			childtab[i].pid = (pid_t)-1;
	}
}
