/*
 * Copyright (C) 1997 WIDE Project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by WIDE Project and
 *	its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/*
 * $Id: plist.c,v 1.12.2.5 1997/12/22 17:03:11 itojun Exp $
 */

#include "mgp.h"

#define	PL_X_MARGIN	10


static int pl_init = 0;
static Pixmap pg_save;
static Pixmap pl_pix;
static Pixmap pl_tpix;
static u_int pg_h, pg_y;

/*
 * Display Page List at the bottom of the window
 */
void
pl_on()
{
	XSetWindowAttributes wattr;
	int i, ny, pl_x;

	if (pl_init == 0) {
		pl_init = 1;
		/* number of pages in x-axis */
		pl_nx = (window_width - 2 * PL_X_MARGIN) / pl_fw / 3;
		/* number of pages in y-axis */
		pl_ny = maxpage / pl_nx + 1;
		/* page list coordinate */
		pl_x = PL_X_MARGIN;
		pl_y = window_height - (pl_ny + 1) * pl_fh;

		for (i = 1; i <= maxpage; i++) {
			ny = (i - 1) / pl_nx;
			plwin[i] = XCreateSimpleWindow(display, window,
				pl_x + 3 * ((i - 1) % pl_nx) * pl_fw,
				pl_y + pl_fh * ny,
				3 * pl_fw, pl_fh, 0,
				BlackPixel(display, 0), WhitePixel(display, 0));
			wattr.do_not_propagate_mask =
				ButtonPressMask|ButtonReleaseMask;
			XChangeWindowAttributes(display, plwin[i],
				CWDontPropagate, &wattr);
			XSelectInput(display, plwin[i],
				ButtonPressMask|EnterWindowMask|
				LeaveWindowMask|StructureNotifyMask);
		}
		pl_pix = XCreatePixmap(display, window, window_width,
				pl_ny * pl_fh, depth);
		pl_tpix = XCreatePixmap(display, window, window_width,
				pl_fh * 2, depth);
	}
	XCopyArea(display, window, pl_pix, gcall, 0, pl_y,
			window_width, pl_ny * pl_fh, 0, 0);
	XMapSubwindows(display, window);
	for (i = 1; i <= maxpage; i++)
		pl_pdraw(i, gc_pl);
	pl_mode = 1;
	XFlush(display);
}

/*
 * Turn off the page list
 */
void
pl_off()
{
	if (pl_mode) {
		XUnmapSubwindows(display, window);
		XCopyArea(display, pl_pix, window, gcall, 0, 0,
			window_width, pl_ny * pl_fh, 0, pl_y);
	}
	pl_mode = 0;
	XFlush(display);
}

/*
 * Draw page numbers in each small window
 * Enbold the current page
 */
void
pl_pdraw(i, gc)
	int i;
	GC gc;
{
	char buf[10];

	sprintf(buf, "%02d", i % 100);
	XDrawImageString(display, plwin[i], gc, (int)(pl_fw / 2),
		(int)(pl_fh / 1.2), buf, 2);
	if (i == cur_page)
		XDrawString(display, plwin[i], gc, (int)(pl_fw / 2) + 1,
			(int)(pl_fh / 1.2), buf, 2);
}

/*
 * Show the title of the page when the mouse enters a window.
 * Turn it off if the page number is zero.
 */
void
pl_title(page)
	int page;
{
	char buf[BUFSIZ];

	if (page == 0) {
		if (pl_tmode)
			XCopyArea(display, pl_tpix, window, gcall, 0, 0,
				window_width, pl_fh * 2, 0, pl_y - pl_fh * 2);
		pl_tmode = 0;
		return;
	}
	pl_tmode = 1;
	XCopyArea(display, window, pl_tpix, gcall, 0, pl_y - pl_fh * 2,
		window_width, pl_fh * 2, 0, 0);
	sprintf(buf, "page %d: %s", page, page_title(page));
	XSetForeground(display, gc_pta, ctrl_color);
	XSetForeground(display, gc_ptk, ctrl_color);
	draw_kstring(window, buf, 10, pl_y - pl_fh / 2);
}

/*
 * Returns a pointer to title of specified page.
 */
char *
page_title(page)
	int page;
{
	int l;
	char *p;
	struct ctrl *cp;

	p = "";
	for (l = 0; l <= page_attribute[page].pg_linenum; l++) {
		cp = page_control[page][l];
		while (cp && cp->ct_op != CTL_TEXT)
			cp = cp->ct_next;
		if (!cp)
			continue;
		if (cp->ctc_value && *cp->ctc_value) {
			p = cp->ctc_value;
			break;
		}
	}

	while (*p && isspace(*p))
		p++;

	return p;
}

void
pg_on(page, showlast)
	u_int page;
	u_int showlast;
{
	char buf[BUFSIZ], *p;
	int n;
	static Pixmap pg_tmp;

	pg_h = pl_fh + plfs->max_bounds.descent;
	pg_y = window_height - pg_h;

	if (pg_tmp == NULL)
		pg_tmp = XCreatePixmap(display, window, window_width, 2 * pl_fh,
			depth);
	if (pg_save == NULL)
		pg_save = XCreatePixmap(display, window, window_width, pg_h,
			depth);
	XCopyArea(display, window, pg_save, gcall, 0, pg_y,
		window_width, pg_h, 0, 0);
	XSetForeground(display, gc_pta, ctrl_color);
	XSetForeground(display, gc_ptk, ctrl_color);
	if (page > 1) {
		sprintf(buf, "\033$B\"+\033(B %s",
			page_title(page - 1));
		draw_kstring(window, buf,
			10, window_height - plfs->max_bounds.descent);
	}
	if (page < maxpage) {
		sprintf(buf, "%s\033$B\"*\033(B",
			page_title(page + 1));
		for (p = buf; *p; p++)
			if (*p == '\n')
				*p = ' ';
		n = draw_kstring(pg_tmp, buf, 0, pl_fh);
		sprintf(buf, "%s\033$B\"*\033(B",
			page_title(page + 1));
		for (p = buf; *p; p++)
			if (*p == '\n')
				*p = ' ';
		draw_kstring(window, buf, window_width - 10 - n,
			window_height - plfs->max_bounds.descent);
	}
	if (showlast) {
		sprintf(buf, "-- %d/%d --", page, maxpage);
		n = draw_kstring(pg_tmp, buf, 0, pl_fh);
		/*
		 * The contents of buf is not destroyed by draw_kstring
		 * as it doesnot include Kanji sequence.
		 */
		draw_kstring(window, buf, window_width / 2 - n / 2,
			window_height - plfs->max_bounds.descent);
	}
}

void
pg_off()
{
	XCopyArea(display, pg_save, window, gcall, 0, 0,
		window_width, pg_h, 0, pg_y);
	XFlush(display);
}

void
pl_config()
{
	int i;

	if (pl_mode)
		pl_off();
	if (pg_mode)
		pg_off();
	if (pl_init) {
		XFreePixmap(display, pl_tpix);
		XFreePixmap(display, pl_pix);
		pl_init = 0;
		for (i = 1; i <= maxpage; i++)
			XDestroyWindow(display, plwin[i]);
	}
	if (pg_save) {
		XFreePixmap(display, pg_save);
		pg_save = NULL;
	}
}

int
draw_kstring(d, buf, x, y)
	Drawable d;
	char *buf;
	int x, y;
{
	char *p;
	int x0 = x;

	p = buf + strlen(buf) - 1;
	if (*p == '\n')
		*p = '\0';
	for (p = buf; *p; p++) {
		if (strncmp(p, "\033$@", 3) == 0 ||
		    strncmp(p, "\033$B", 3) == 0) {
			*p = '\0'; p += 3;
			/* draw 8bit char */
			XDrawString(display, d, gc_pta, x, y,
				buf, strlen(buf));
			x += XTextWidth(plfs, buf, strlen(buf));
			buf = p--;
		} else if (strncmp(p, "\033(B", 3) == 0 ||
			   strncmp(p, "\033(J", 3) == 0) {
			*p = '\0'; p += 3;
			/* draw 16bit char */
			XDrawString16(display, d, gc_ptk, x, y,
				(XChar2b *)buf, strlen(buf)/2);
			x += XTextWidth16(plkfs, (XChar2b *)buf, strlen(buf)/2);
			buf = p--;
		}
	}
	/* draw 8bit char */
	if (*buf != '\0') {	/* Assumed ASCII at the end of string */
		XDrawString(display, d, gc_pta, x, y, buf, strlen(buf));
		x += XTextWidth(plfs, buf, strlen(buf));
	}
	return x - x0;
}
