/*
 * Copyright (C) 1997 WIDE Project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by WIDE Project and
 *	its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/*
 * $Id: print.c,v 1.19.2.16 1997/12/26 05:52:41 itojun Exp $
 */

#include "mgp.h"

static u_int align = AL_LEFT;
static u_char *curprefix = NULL;
static u_char *tabprefix = NULL;
static u_int leftfillpos = 0;

#define A4_WIDTH	760
#define A4_HEIGHT	590
#define A4_XMARGIN	10
#define A4_YMARGIN	5
#define PRINT_ASCIIFONT		"Times-Roman"
#define PRINT_ASCIITHICKFONT	"Helvetica"
#define PRINT_KANJIFONT		"Ryumin-Light-H"
#define PRINT_KANJITHICKFONT	"GothicBBB-Medium-H"

static FILE *fp;
static u_int pos_x;
static u_int pos_y;
static u_int last_pos_x;
static u_int last_pos_y;
static u_int cont;
static u_int string_width;
static u_int reverse;
static u_char outputfile[MAXVALLEN];
static u_char *prefix = NULL;

static int needstart = 0;
static int needclose = 0;

/* the order must be the same for the following two decls. */
#define SPACE	0
#define ASCII	1
#define KANJI	2
#define NOPE	-1
static int print_font_bold[] = {
	0, 0, 0, 0,
};
static char *print_font_string[] = {
	PRINT_ASCIIFONT, PRINT_ASCIITHICKFONT,
	PRINT_ASCIIFONT, PRINT_ASCIITHICKFONT,
	PRINT_KANJIFONT, PRINT_KANJITHICKFONT,
	NULL,
};
static char *print_font_abbrev[] = {
	"SPC", "SPCT", "ASC", "ASCT", "KNJ", "KNJT", NULL,
};

static void print_out __P((void));
static void print_page __P((u_int));
static void print_init __P((void));
static void print_pageinit __P((u_int));
static void process_direc_print __P((struct ctrl *));
static void print_bar __P((struct ctrl *));
static void linestart __P((void));
static void print_line __P((u_char *));
static void print_string __P((u_char *));
static u_int print_strlen __P((u_char *));
static void image_load_print __P((char *, int, int, int, int, int));
static void print_usage __P((char *));
static int set_position __P((void));
static void print_full_image __P((XImage *, Visual *));
static void print_256_image __P((XImage *));
static void print_eps __P((FILE *, char *, int, int, int, int, int));

/* image/ *.[ch] */
extern char *Paths[];
extern int NumPaths;
extern char *expandPath __P((char *));
extern void bestVisual __P((Display *, int, Image *, Visual **, unsigned int *));

extern int findImage __P((char *, char *));

static void
print_out()
{
	u_int	i;
	u_int	width, height;

	width  = window_width;
	height = window_height;
	window_width  = A4_WIDTH;
	window_height = A4_HEIGHT;

	if (outputfile[0]) {
		if ((fp = fopen(outputfile, "w")) == NULL) {
			perror("fopen");
			exit(-1);
		}
	} else
		fp = stdout;

	print_init();
	for (i = 1; i <= maxpage; i ++) {
		print_page(i);
	}

	fclose(fp);

	window_width  = width;
	window_height = height;
}

static void
print_page(page)
	u_int page;
{
	u_int line;
	struct ctrl *cp;

	pos_x = pos_y = 0;
	print_pageinit(page);

	for (line = 0;
	     line <= page_attribute[page].pg_linenum;
	     line++) {
		prefix = 0;

#if 0
		/* process default control */
		if (!(page_attribute[page].pg_flag & PGFLAG_NODEF))
			for (cp = default_control[line]; cp; cp = cp->ct_next)
				process_direc_print(cp);
#endif


		/* process normal control */
		for (cp = page_control[page][line]; cp; cp = cp->ct_next)
			process_direc_print(cp);
	}

	fprintf(fp, "showpage\n\n");
}

static void
print_init()
{
	char **p;
	char **q;

	fprintf(fp, "%%!PS-Adobe-2.0 EPSF-2.0\n");
	fprintf(fp, "%%%%Creater: mgp2ps\n");
	fprintf(fp, "%%%%Title: mgp2ps output file\n");
	fprintf(fp, "%%%%Pages: %d\n", maxpage);
	fprintf(fp, "%%%%BoundingBox: 0 0 612 792\n");
	fprintf(fp, "%%%%DocumentPaperSizes: A4\n");
	fprintf(fp, "%%%%Orientation: Landscape\n");
	fprintf(fp, "%%%%EndComments\n");

	/* define writebox */
	fprintf(fp, "/writebox { %d %d moveto\n", A4_XMARGIN, A4_YMARGIN * -1);
	fprintf(fp, "%d %d rlineto\n", 0, A4_HEIGHT * -1);
	fprintf(fp, "%d %d rlineto\n", A4_WIDTH, 0);
	fprintf(fp, "%d %d rlineto\n", 0, A4_HEIGHT);
	fprintf(fp, "%d %d rlineto stroke } def \n", A4_WIDTH * -1, 0);

	for (p = print_font_string, q = print_font_abbrev;
	     *p && *q;
	     p++, q++) {
		fprintf(fp, "/%s {/%s findfont exch scalefont setfont} def\n",
			*q, *p);
	}

	/* a spell for EPS */
	fprintf(fp, "%%\n");
	fprintf(fp, "/BeginEPSF {%%def\n");
	fprintf(fp, "  /b4_Inc_state save def\n");
	fprintf(fp, "  /dict_count countdictstack def\n");
	fprintf(fp, "  /op_count count 1 sub def\n");
	fprintf(fp, "  userdict begin\n");
	fprintf(fp, "  /showpage {}def\n");
	fprintf(fp, "  0 setgray 0 setlinecap\n");
	fprintf(fp, "  1 setlinewidth 0 setlinejoin\n");
	fprintf(fp, "  10 setmiterlimit [] 0 setdash\n");
	fprintf(fp, "  newpath\n");
	fprintf(fp, "  /languagelevel where\n");
	fprintf(fp, "  {pop languagelevel\n");
	fprintf(fp, "  1 ne\n");
	fprintf(fp, "    {false setstrokeadjust\n");
	fprintf(fp, "     false setoverprint\n");
	fprintf(fp, "    }if\n");
	fprintf(fp, "  }if\n");
	fprintf(fp, "}bind def\n");
	fprintf(fp, "%%\n");
	fprintf(fp, "/EndEPSF {%%def\n");
	fprintf(fp, "  count op_count sub {pop}repeat\n");
	fprintf(fp, "  countdictstack\n");
	fprintf(fp, "  dict_count sub {end}repeat\n");
	fprintf(fp, "  b4_Inc_state restore\n");
	fprintf(fp, "}bind def\n");
	fprintf(fp, "%%\n");
}

static void
print_pageinit(page)
	u_int page;
{
	fprintf(fp, "%%%%Page: %d %d\n", page, page);
	fprintf(fp, "90 rotate newpath writebox\n");
}

static void
process_direc_print(cp)
	struct ctrl *cp;
{
	switch(cp->ct_op) {
	case CTL_SIZE:
		char_size = window_height * cp->cti_value / 100;
		break;

#ifdef VFLIB
	case CTL_VFONT:
		print_font_bold[KANJI] = strstr(cp->ctc_value, "goth") ? 1 : 0;
		break;
#endif /*VFLIB*/

	case CTL_XFONT:
		print_font_bold[SPACE] = print_font_bold[ASCII] =
			strstr(cp->ctc_value, "helvetica") ? 1 : 0;
		break;

	case CTL_VGAP:
		vert_gap = cp->cti_value;
		break;

	case CTL_HGAP:
		horiz_gap = cp->cti_value;
		break;

	case CTL_CENTER:
		align = AL_CENTER;
		break;

	case CTL_LEFT:
	case CTL_LEFTFILL:	/*XXX*/
		align = AL_LEFT;
		break;

#if 0 /*yet to be done*/
	case CTL_LEFTFILL:
		align = AL_LEFTFILL0;
		break;
#endif

	case CTL_RIGHT:
		align = AL_RIGHT;
		break;

	case CTL_CONT:
		cont = 1;
		break;

	case CTL_IMAGE:
		needstart = 0;
		image_load_print(cp->ctm_fname, cp->ctm_numcolor,
			cp->ctm_ximagesize, cp->ctm_yimagesize, 0,
			cp->ctm_zoomflag);
		break;

	case CTL_BAR:
		print_bar(cp);
		break;

	case CTL_PREFIX:
		curprefix = cp->ctc_value;
		break;

	case CTL_TABPREFIX:
		tabprefix = cp->ctc_value;
		break;

	case CTL_PREFIXPOS:
		if (tabprefix)
			print_line(tabprefix);
		else if (curprefix)
			print_line(tabprefix);
		break;

	case CTL_TEXT:
		if (!cp->ctc_value)
			break;
#if 0
		if (align == AL_LEFTFILL0) {
			leftfillpos = string_width;
			align = AL_LEFTFILL1;
		}
#endif
		print_line(cp->ctc_value);
		break;

	case CTL_LINESTART:
		linestart();
		fprintf(fp, "/tmp {} def\n");
		string_width = 0;
		needstart = 1;
		break;

	case CTL_LINEEND:
		pos_x = set_position();
		if (needclose) {
			fprintf(fp, "} def\n");
			needclose = 0;
		}
		if (!cont) {
			fprintf(fp, "%d %d moveto tmp\n", pos_x + A4_XMARGIN,
				-VERT_POS(pos_y, char_size) - A4_YMARGIN - char_size);
		}
		pos_y += VERT_STEP(char_size);
		tabprefix = NULL;
#if 0
		if (align == AL_LEFTFILL1) {
			align = AL_LEFTFILL0;
			leftfillpos = 0;
		}
#endif
		break;

	default:
		break;
	}
}

static void
print_bar(cp)
	struct ctrl *cp;
{
	struct ctrl_bar pbar;

	pbar.ct_width = cp->ctb_width * A4_HEIGHT / 1000;
	pbar.ct_start = cp->ctb_start * A4_WIDTH / 100;
	pbar.ct_length = cp->ctb_length * A4_WIDTH / 100;

	fprintf(fp, "%%bar color %d %d %d\n",
		cp->ctb_width, cp->ctb_start, cp->ctb_length);
	fprintf(fp, "%d %d moveto\n",
		A4_XMARGIN + pbar.ct_start,
		pos_y * -1 - A4_YMARGIN);
	fprintf(fp, "%d %d rlineto\n", 0, pbar.ct_width * -1);
	fprintf(fp, "%d %d rlineto\n", pbar.ct_length, 0);
	fprintf(fp, "%d %d rlineto\n", 0, pbar.ct_width);
	fprintf(fp, "%d %d rlineto stroke\n", pbar.ct_length * -1, 0);

	last_pos_y = pos_y;
	last_pos_x = cp->ctb_start + cp->ctb_length;
	pos_y += cp->ctb_width + VERT_GAP(char_size) / 2;
}

static void
linestart()
{
}

static void
print_line(data)
	u_char *data;
{
	if (needstart) {
		fprintf(fp, "/tmp {\n");
		needstart = 0;
		needclose = 1;
	}

	if (data)
		string_width += print_strlen(data);

	/* skip */
	if (data && *data)
		print_string(data);

	last_pos_x = pos_x + string_width;
	last_pos_y = pos_y;

	cont = 0;
}

static void
print_string(data)
	u_char *data;
{
	int ctype, last_ctype;
	u_char *p;
	u_int last_size = 0;
	u_int code;
#define OPENBRACE(ctype)	((ctype) == KANJI ? '<' : '(')
#define CLOSEBRACE(ctype)	((ctype) == KANJI ? '>' : ')')

	p = data;
	ctype = ASCII;
	last_ctype = NOPE;

	while (*p != '\n' && *p != 0) {
		if (*p == 0x1b && *(p+1) == '$' &&
				(*(p+2) == 'B' || *(p+2) == '@')) {
			ctype = KANJI;
			p += 3;
			continue;
		}
		if (*p == 0x1b && *(p+1) == '(' &&
				(*(p+2) == 'B' || *(p+2) == 'J')) {
			ctype = ASCII;
			p += 3;
			continue;
		}
		if (ctype != KANJI) {
			if (isspace(*p))
				ctype = SPACE;
			else
				ctype = ASCII;
		}

		if (last_size != char_size
		 || last_ctype != ctype
		 || strcmp(print_font_string[last_ctype * 2],
				 print_font_string[ctype * 2]) != 0) {
			int idx;

			/* we need to change font */
			if (last_ctype != NOPE) {
				fprintf(fp, "%c show\n",
					CLOSEBRACE(last_ctype));
			}

			idx = ctype * 2 + (print_font_bold[ctype] ? 1 : 0);
			fprintf(fp, "%d %s ", char_size,
				print_font_abbrev[idx]);

			fprintf(fp, "%c", OPENBRACE(ctype));
		}

		if (ctype == KANJI) {
			code = *p * 256 + *(p + 1);
			fprintf(fp, "%04x ", code);
			p += 2;
		} else {
			if ((*p & 0xff) == OPENBRACE(ctype)
			 || (*p & 0xff) == CLOSEBRACE(ctype)) {
				fprintf(fp, "%c show <%x> show %c",
					CLOSEBRACE(ctype), *p & 0xff,
					OPENBRACE(ctype));
			} else
				fprintf(fp, "%c", *p & 0xff);
			p ++;
		}
		last_ctype = ctype;
		last_size = char_size;
	}

	if (last_ctype != NOPE)
		fprintf(fp, "%c show\n", CLOSEBRACE(last_ctype));
}

static u_int
print_strlen(data)
	u_char *data;
{
	u_char *p;
	u_int kanji = 0;
	u_int code;
	u_int length;

	p = data;
	length = 0;

	while (*p != '\n' && *p != 0) {
		if (*p == 0x1b && *(p+1) == '$' &&
				(*(p+2) == 'B' || *(p+2) == '@')) {
			kanji = 1;
			p += 3;
			continue;
		}
		if (*p == 0x1b && *(p+1) == '(' &&
				(*(p+2) == 'B' || *(p+2) == 'J')) {
			kanji = 0;
			p += 3;
			continue;
		}

		if (kanji) {
			code = *p * 256 + *(p + 1);
			p += 2;
		} else {
			length += char_size/2;
			p ++;
			continue;
		}

		length += char_size;
	}

	return length;
}

static void
image_load_print(filename, numcolor, ximagesize, yimagesize, backflag, zoomflag)
	char *filename;
	int numcolor;
	int ximagesize;
	int yimagesize;
	int backflag;
	int zoomflag;
{
	Image *image, *myimage;
	Pixmap mypixmap;
	XImageInfo *ximageinfo;
	XImage	*print_image;
	Visual *visual;
	u_int depth;
	int width, height;
	int xzoomrate, yzoomrate;
	int	verbose = mgp_flag & FL_VERBOSE;
	static Cursor curs;
	u_int	print_width, print_height;

	if (zoomflag == 2)
		ximagesize = yimagesize = 0;

	/*
	 * hook for eps file.
	 */
    {
	char epsfile[MAXPATHLEN], *p;
	char fullname[MAXPATHLEN];
	FILE *epsfp;

	strcpy(epsfile, filename);

	p = strrchr(epsfile, '.');
	if (p) {
		*p = '\0';
	}
	strcat(epsfile, ".eps");

	fullname[0] = '\0';
	if (findImage(epsfile, fullname) == 0) {
		epsfp = fopen(fullname, "r");
		print_eps(epsfp, fullname, numcolor, ximagesize, yimagesize,
			backflag, zoomflag);
		fclose(epsfp);
		return;
	}
    }

	if (!curs)
		curs = XCreateFontCursor(display, XC_watch);
	XDefineCursor(display, window, curs); XFlush(display);

	screen = DefaultScreen(display);

	if ((myimage = loadImage(filename, verbose)) == NULL)
		exit(-1);	/* fail to load image data */
	width = myimage->width;
	height = myimage->height;

	if (numcolor)
		myimage = reduce(myimage, numcolor, verbose);

	if (ximagesize != 0 && yimagesize != 0) {
		if (!zoomflag) {
			xzoomrate = window_width * ximagesize / width;
			yzoomrate = window_height * yimagesize / height;
		} else {
			xzoomrate = ximagesize;
			yzoomrate = yimagesize;
		}
		image = myimage;
		myimage = zoom(image, xzoomrate, yzoomrate, verbose);
		freeImage(image);
	} else
		xzoomrate = yzoomrate = 100;

	bestVisual(display, screen, myimage, &visual, &depth);
	if (verbose && (visual != DefaultVisual(display, screen)))
		printf("  Using %s visual\n", nameOfVisualClass(visual->class));

	if (! (ximageinfo= imageToXImage(display, screen, visual, depth, myimage,
		0, 0, verbose))) {
			fprintf(stderr, "Cannot convert Image to XImage\n");
			exit(1);
	}

	if (((mypixmap= ximageToPixmap(display,
		RootWindow(display, 0), ximageinfo)) == None) && verbose) {
			fprintf(stderr, "Cannot create image in server\n");
			exit(1);
	}

	switch (align) {
	case AL_LEFT:
	case AL_LEFTFILL0:
		pos_x = 0;
		break;
	case AL_LEFTFILL1:
		pos_x = leftfillpos;
		break;
	case AL_CENTER:
		pos_x = (window_width - width * xzoomrate /100) / 2;
		break;
	case AL_RIGHT:
		pos_x = window_width - width * xzoomrate /100;
		break;
	}

	if (cont) {
		pos_x = last_pos_x;
		pos_y = last_pos_y;
		cont = 0;
	}

	print_width = myimage->width;
	print_height = myimage->height;

	print_image =  XGetImage(display, mypixmap, 0, 0,
					print_width, print_height,
					AllPlanes, ZPixmap);

	fprintf(fp, "gsave\n");
	fprintf(fp, "%d %d translate\n",pos_x + A4_XMARGIN,
		(pos_y + print_height + A4_YMARGIN + VERT_GAP(char_size)) * -1);
	fprintf(fp, "%d %d scale\n", print_width, print_height);
	fprintf(fp, "%d %d 8\n", print_width, print_height);
	fprintf(fp, "[%d 0 0 %d 0 %d]\n", print_width,
			-1 * print_height, print_height);
	fprintf(fp, "{currentfile\n");
	fprintf(fp, "%d string readhexstring pop}\n", print_width * 3);
	fprintf(fp, "false 3\n");
	fprintf(fp, "colorimage\n");

	if (print_image->bits_per_pixel == 8) {
			print_256_image(print_image);
	 }else {
		if (print_image->bits_per_pixel == 16) {
			print_full_image(print_image, visual);
		} else {
			fprintf(stderr, "Sorry unsupported visual %d\n",
				print_image->bits_per_pixel);
			exit(-1);
		}
	}

	fprintf(fp, "grestore\n");

	last_pos_y = pos_y;
	last_pos_x = pos_x + width * xzoomrate / 100;
	pos_y += height * yzoomrate / 100 + VERT_GAP(char_size);

	XFreePixmap(display, mypixmap);
	freeXImage(myimage, ximageinfo);
	freeImage(myimage);
	XUndefineCursor(display, window); XFlush(display);
}

static void
print_usage(name)
	char	*name;
{
	fprintf(stderr, "Usage: %s [-rV] [-f psfile] mgpfile\n", name);
	exit(0);
}

static int
set_position()
{
	u_int   x;

	switch(align) {

	case AL_CENTER:
		x = (window_width - string_width)/ 2;
		break;

	case AL_LEFT:
	case AL_LEFTFILL0:
	case AL_LEFTFILL1:
		x = 0;
		break;

	case AL_RIGHT:
		x = window_width - string_width;
		break;

	default:
		x = 0;
		break;
	}

	return x;
}

static void
print_full_image(print_image, visual)
	XImage	*print_image;
	Visual	*visual;
{
	u_int	x, y;
	u_int	r, g, b;
	u_int	count = 0;
	u_int	width, height, byte_width;
	u_short	*data;

	data = (u_short *)print_image->data;
	width = print_image->width;
	height = print_image->height;
	byte_width = print_image->bytes_per_line / sizeof(u_short);

	for (y = 0; y < height; y ++) {
		for (x = 0; x < width; x ++) {
			r = ((data[x + y * byte_width] & visual->red_mask) >> 11) & 0xff;
			g = ((data[x + y * byte_width] & visual->green_mask) >> 5) & 0xff;
			b = (data[x + y * byte_width] & visual->blue_mask) & 0xff;

			if (reverse)
				fprintf(fp, "%02x%02x%02x",
					~(r << 3) & 0xff, ~(g << 2) & 0xff, ~(b << 3) & 0xff);
			else
				fprintf(fp, "%02x%02x%02x", r << 3, g << 2, b << 3);

			count ++;
			if (count == 10) {
				count = 0;
				fprintf(fp, "\n");
			}
		}
	}
}

static void
print_256_image(print_image)
	XImage	*print_image;
{
	u_int	x, y;
	u_int	count = 0;
	u_char	*data;
	u_int	width, height, byte_width;

	data = (u_char *)print_image->data;
	width = print_image->width;
	height = print_image->height;
	byte_width = print_image->bytes_per_line / sizeof(u_char);

	for (y = 0; y < height; y ++) {
		for (x = 0; x < width; x ++) {
			if (reverse)
				fprintf(fp, "%02x%02x%02x",
					~data[x + y * byte_width] & 0xff,
					~data[x + y * byte_width] & 0xff,
					~data[x + y * byte_width] & 0xff);
			else
				fprintf(fp, "%02x%02x%02x",
					data[x + y * byte_width],
					data[x + y * byte_width],
					data[x + y * byte_width]);

			count ++;
			if (count == 10) {
				count = 0;
				fprintf(fp, "\n");
			}
		}
	}
}

#define PSSCALEF	1.25
#define PSSCALE		(100.0 / 80.0)
#define PSBASE		(80.0 / 100.0)

static void
print_eps(epsfp, epsfile, numcolor, ximagesize, yimagesize, backflag, zoomflag)
	FILE *epsfp;
	char *epsfile;
	int numcolor;
	int ximagesize;
	int yimagesize;
	int backflag;
	int zoomflag;

{
	char filename[MAXPATHLEN];
	char line1[BUFSIZ];
	char line2[BUFSIZ];
	char line[BUFSIZ];
	int x1, y1, x2, y2, height, width, sheight, swidth;

	/*
	 * mgp scale is NOT supported.
	 *
	 * PS scale code is for visibility when printed.
	 * Kazu's epsfile should be enlarged 1.2 times.
	 * And this is a challenge to understand PS's scale...
	 */
	if (fgets(line1, sizeof(line1), epsfp) == NULL) {
		fprintf(stderr, "no first line in %s.\n", filename);
		exit(1);
	}
	if (strncmp(line1, "%!", 2) != 0) {
		fprintf(stderr, "non eps file %s used as eps.\n", filename);
		exit(1);
	}

	if (fgets(line2, sizeof(line2), epsfp) == NULL) {
		fprintf(stderr, "no second line in %s.\n", filename);
		exit(1);
	}

	sscanf(line2, "%%%%BoundingBox: %d %d %d %d", &x1, &y1, &x2, &y2);

	width = x2 - x1;
	height = y2 - y1;
	swidth = (int)(width * PSSCALE);
	sheight = (int)(height * PSSCALE);

	switch (align) {
	case AL_LEFT:
		pos_x = 0;
		break;
	case AL_CENTER:
		pos_x = (window_width - swidth) / 2;
		break;
	case AL_RIGHT:
		pos_x = window_width - swidth;
		break;
	}

	if (cont) {
		pos_x = last_pos_x;
		pos_y = last_pos_y;
		cont = 0;
	}

	fprintf(fp, "BeginEPSF\n");
	fprintf(fp, "%f %f scale\n", PSSCALEF, PSSCALEF);
	fprintf(fp, "%d %d translate\n",
		(x1 - (int)(A4_XMARGIN * PSBASE)) * -1,
		(y2 + (int)(A4_YMARGIN * PSBASE)) * -1);
	fprintf(fp, "%d %d translate\n",
		(int)(pos_x * PSBASE),
		(int)(pos_y * PSBASE * -1));
	fprintf(fp, "%%BeginDocument: %s\n", epsfile);
	fputs(line1, fp);
	fputs(line2, fp);
	while (fgets(line, sizeof(line), epsfp)) {
		if (strncmp(line, "%%PageBoundingBox:", 18) == 0)
			line[1] = '#';
		fputs(line, fp);
	}
	fprintf(fp, "%%EndDocument\n");
	fprintf(fp, "EndEPSF\n");

	last_pos_y = pos_y;
	last_pos_x = pos_x + swidth;
	pos_y += sheight + VERT_GAP(char_size);
}

/*------------------------------------------------------------*/

int
main(argc, argv)
	int argc;
	char **argv;
{
	u_int opt;
	extern char *optarg;
	extern int optind;
	u_char *progname;
	static char pathbuf[MAXPATHLEN];

	progname = argv[0];

	while ((opt = getopt(argc, argv, "Vrf:")) != -1) {
		switch (opt) {
		case 'f':
			strcpy(outputfile, optarg);
			break;

		case 'r' :
			reverse = 1;
			break;

		case 'V' :
			mgp_flag |= FL_VERBOSE;
			break;

		default:
			print_usage(progname);
			/*NOTREACHED*/
		}
	}

	argc -= optind;
	argv += optind;

	if (argc != 1) {
		print_usage(progname);
		/*NOTREACHED*/
	}

	mgp_fname = argv[0];

	/* setting up the search path. */
    {
	char *p;
	loadPathsAndExts();
	strncpy(pathbuf, mgp_fname, sizeof(pathbuf));
	if (p = rindex(pathbuf, '/')) {
		*p = '\0';
        	Paths[NumPaths++]= expandPath(pathbuf);
	}
    }

	init_win1();
	load_file(mgp_fname);
	init_win2();
	print_out();

	exit(0);
}
